
require 'rexml/document'
require 'enumerator'
require 'amrita2/util'

module REXML
  class Child 
    attr_accessor :amrita_no_spec
    def amrita_compile_main_code(cg)
      #cg.put_constant(to_s)
      cg.put_rexml_node(self)
    end
  end

  class Parent < Child
    def amrita_compile_main_code(cg)
      if amrita_no_spec
        super
      else
        each_child do |c|
          c.amrita_compile_main_code(cg)
        end
      end
    end
  end

  class DocType < Parent 
    def amrita_compile_main_code(cg)
      cg.put_constant(to_s)
      cg.put_constant("\n")
      #cg.put_rexml_node(self)
    end
  end

  class Comment < Child
    def amrita_compile_main_code(cg)
      cg.put_constant("<!--#{ to_s}-->")
    end
  end

  class Element < Parent
    attr_accessor :amrita_spec
    def amrita_compile_main_code(cg)
      if amrita_no_spec
        super
      else
        cg.compile_element_main_code(self) do
          super
        end
      end
    end

    def amrita_compile_sub_code(cg)
      cg.compile_element_sub_code(self) do
        elements.each do |e|
          e.amrita_compile_sub_code(cg)
        end
      end
    end

     alias to_s_org to_s
     def to_s
       Amrita2::SanitizedString[to_s_org]
     end
  end
end

module Amrita2
  module Core
    class CodeGenerator
      attr_accessor :current_stream, :ie_hack 

      def initialize(stream = "get_mainstream")
        @current_stream = stream
        init_results
        init_strbuf
        @current_indent = 0
        @varcnt=0
        @ie_hack = false
      end

      def put_rexml_node(node)
        s = ""
        node.write(s, -1, false, @ie_hack)
        put_constant(s)
      end

      def put_constant(msg)
        @strbuf.concat msg
      end

      def flush
        if @strbuf.size > 0
          @results << [@current_indent, "#{current_stream} << " + @strbuf.inspect]
          init_strbuf
        end
      end

      def put_expression(exp)
        code("#{current_stream} << #{exp}")
      end

      def code(l)
        flush
        @results << [@current_indent, l]
      end

      def new_var
        @varcnt += 1
        "v#{@varcnt}"
      end

      def results
        flush
        @results.collect do |level, line|
          "  " * level + line
        end
      end

      def if_(cond)
        code("if #{cond}")
        level_up do
          yield
        end
        code("end")
      end

      def else_
        level_down do
          code("else")
        end
        yield
      end

      def level_up
        flush
        @current_indent += 1
        yield
      ensure
        flush
        @current_indent -= 1
      end

      def level_down
        flush
        @current_indent -= 1
        yield
      ensure
        flush
        @current_indent += 1
      end

      def define_method(name, *params)
        if params.size == 0
          code("def #{name}")
        else
          code("def #{name}(#{params.join(',')})")
        end
        level_up do
          yield
        end
        code("end")
      end

      def define_element_method(name, new_stream, &block)
        define_method(name, "val=nil", "attrs={}", "&block") do
          code("attrs = val if val.kind_of?(Hash)")
          new_context("get_substream(:#{new_stream})") do
            yield
          end
        end
        code("module_function #{name.inspect}")
      end

      def define_module(name)
        code("module #{name}")
        level_up do
          yield
        end
        code("end")
      end

      def convert(encto, encfrom, *args)
        a = args.join(",")
        code("#{a} = *Iconv::iconv(#{encto.inspect}, #{encfrom.inspect}, #{a})")
      end

      def convert_hash(encto, encfrom, hash)
        code("#{hash}.each do |k, v|")
        level_up do
          code("v = *Iconv::iconv(#{encto.inspect}, #{encfrom.inspect}, v)")
          code("#{hash}[k] = v")
        end
        code("end")
      end

      def compile_element_main_code(element, &block)
        spec = element.amrita_spec
        if spec
          spec.compile_element_main_code(element, self, &block)
        else
          generate_static_element_code(element, &block)
        end
      end

      def compile_element_sub_code(element, &block)
        spec = element.amrita_spec
        if spec
          spec.compile_element_sub_code(element, self, &block) 
        else
          yield
        end
      end

      def init_results
        @results = []
      end

      def init_strbuf
        @strbuf = ""
      end

      def generate_static_element_code(e, option={}, &block)
        if e.amrita_no_spec
          put_rexml_node(e)
        elsif e.name and e.name.size > 0
          unless option[:delete_tag]
            a = []
            a << e.name 
            if (option[:delete_id])
              a += e.attributes.enum_for(:each_attribute).find_all do |attr|
                attr.name != option[:key_attr_name].to_s
              end.collect do |attr| 
                attr.to_string 
              end
            else
              a += e.attributes.enum_for(:each_attribute).collect { |attr| attr.to_string }
            end
            put_constant("<#{a.join(' ')}>")
          end
          yield
          put_constant("</#{e.name}>") unless option[:delete_tag]
        else
          yield
        end
      end

      def new_context(new_stream,  &block)
        code("new_context(#{new_stream}) do")
        level_up do
          yield
        end
        code("end")
      end

      #       def new_context(new_stream,  &block)
      #         save_stream = @current_stream
      #         code("new_context(#{new_stream}) do")
      #         var = new_var
      #         code("#{var} = get_mainstream")
      #         @current_stream = var
      #         level_up do
      #           yield
      #         end
      #         code("end")
      #         @current_stream = save_stream
      #       end
    end

    class Matcher
      def find(e)
        return e if match(e)
        e.elements.each do |c|
          x = find(c)
          return x if x
        end
        nil
      end
    end

    class AttrMatcher < Matcher
      attr_reader :sym
      def initialize(sym, val)
        @sym = sym
        @sym_s = sym.to_s
        @val = val
      end

      def match(e)
        if @val != '*'
          @val == e.attributes[@sym_s]
        else
          e.attributes[@sym_s]
        end
      end
    end

    class IDMatcher < AttrMatcher
      def initialize(val)
        super(:id, val.to_s)
      end
    end

    class TagMatcher < Matcher
      def initialize(tag)
        @tag = tag.to_s
      end

      def match(e)
        e.name == @tag
      end
    end

    class AnyMatcher < Matcher
      def match(e)
        true
      end
    end

    module Runtime
      def new_element(tag, attrs={})
        ret = REXML::Element.new(tag.to_s)
        attrs.each do |k,v|
          ret.attributes[k.to_s] = v.to_s
        end
        ret
      end
      
      def start_tag(e, out="")
        out << "<#{e.expanded_name}"
        e.attributes.each_attribute do |attr|
          next if attr.value == ""
          out << " "
          attr.write(out)  
        end unless e.attributes.empty?
        out << ">"
        out
      end

      def end_tag(e, out="")
        out << "</#{e.expanded_name}>"
        out
      end

      class Context 
        attr_reader :main_stream
        attr_accessor :data
        
        def initialize(stream)
          @main_stream = stream
          @sub_stream = {}
        end

        def sub_stream(key)
          @sub_stream[key] ||= ""
        end

        def clear_substream(key)
          @sub_stream[key] = ""
        end
      end
      
      def context_stack
        Thread::current[:amrita_context_stack] ||= []
      end

      def current_context
        context_stack.first
      end

      def new_context(stream="", &block)
        context_stack.unshift Context.new(stream)
        yield
        if stream.kind_of?(String)
          SanitizedString[stream]
        else
          stream
        end
      ensure
        context_stack.shift
      end

      def get_mainstream
        current_context.main_stream
      end

      def get_substream(key)
        current_context.sub_stream(key)
      end

      def output_substream(key)
        c = current_context
        s = c.sub_stream(key)
        c.main_stream << SanitizedString[s]
        c.clear_substream(key)
      end
    end

    class SpecOptionMeta
      module MetaData
        def check_option(spec, k, v)
        end
      end

      @@options = {}
      def self.get_option_meta(name)
        @@options[name]
      end

      def self.define_option(name, meta=nil, *args)
        case meta
        when nil
          meta = Object.new
          meta.extend MetaData
        when MetaData
          # do nothing
        when Module
          m = Object.new
          m.extend meta
          meta = m
        end
        args.each do |mod|
          meta.extend mod
        end
        @@options[name] = meta
      end

      def self.check_options(spec)
        spec.option.each do |k, v|
          meta = @@options[k]
          case meta
          when nil
            raise NameError, "option #{k} was not defined"
          else
            meta.check_option(spec, k, v)
          end
        end
      end

      def self.setup_compiler(spec, compiler)
        spec.option.each do |k, v|
          next unless v
          meta = @@options[k]
          case meta
          when CompilerOption
            compiler.extend meta 
          end
        end
      end
      
      def self.inherit_option(spec, child_opt)
        spec.option.each do |k, v|
          meta = @@options[k]
          case meta
          when InheritableOption
            meta.inherite_option(k, v, child_opt)
          end
        end
      end

      def self.define_element_ext(spec, name, sym, params, &block)
        meta = @@options[name]
        if meta 
          meta.define_element_ext(spec, name, sym, params || {}, &block)
        end
      end

      module SymbolOption
        def check_option(spec, k, v)
          case v
          when Symbol, String
          else
            raise "option #{k} must be a Symbol" 
          end
        end
      end

      module InheritableOption
        include MetaData
        def inherite_option(k, v, child_opt)
          return if child_opt.has_key?(k)
          child_opt[k] = v if v
        end
      end

      module CompilerOption
        include MetaData
      end

      module ElementExt
        include MetaData
        def define_element_ext(spec, name, sym, params, &block)
          pp = params.clone 
          pp[name] = true
          spec.dynamic_element(sym, pp, &block)
        end
      end

      define_option(:direct_define, InheritableOption)
      define_option(:key_attr_name, SymbolOption, InheritableOption)
      define_option(:matcher)
      define_option(:delete_tag)
      define_option(:output_stream, SymbolOption)
      define_option(:delete_id, InheritableOption)
      define_option(:place_holder)
      define_option(:match_many)
      define_option(:generate_template, InheritableOption)
      define_option(:encoding, InheritableOption)
      define_option(:dummy_element, ElementExt)
      define_option(:parts_element, ElementExt)
      define_option(:ie_hack)
      define_option(:debug_source)
    end

    class DynamicElementSpec
      attr_accessor :cloned, :parent
      attr_reader :sym, :matcher, :children, :option, :assigned_elements

      def initialize(sym, parent, option={}, &block)
        case sym
        when Symbol
          @sym = sym
        when nil
          @sym = nil
        else
          @sym = sym.to_s.intern
        end
        @parent = parent
        @cloned = false
        @option, @children = option, []

        SpecOptionMeta::check_options(self)
        @direct_define = @option[:direct_define]
        setup_matcher
        setup_compiler
        if block_given?
          @in_setup = true
          self.instance_eval(&block) 
          @in_setup = false
        end

        check_double_id
      end

      def inspect(lvl=0)
        spc = "  " * lvl
        if children.size > 0
          "#{spc}dynamic_element(#{sym.inspect}, #{@option.inspect}, #{self.class.name}) do\n" +
            children.collect {|c| c.inspect(lvl+1)} .join +
            "#{spc}end\n" 
        else
          "#{spc}dynamic_element(#{sym.inspect}, #{@option.inspect}, #{self.class.name})\n"
        end
      end

      def clone
        new_spec = self.class.new(sym, nil, option)
        @children.each_with_index do |child, i|
          c = child.clone
          c.parent = new_spec
          new_spec.children << c
        end
        new_spec.cloned = true
        new_spec
      end

      def assign_element(e)
        raise "can't assign not cloned dynamic spec" unless @cloned
        @assigned_elements ||= []
        @assigned_elements << e
      end

      def matched_many?
        @option[:match_many] and @assigned_elements.size > 1 
      end

      def first_match?(element)
        matched_many? and @assigned_elements.index(element) == 0
      end

      def name_for(element)
        if matched_many?
          i = @assigned_elements.index(element)
          "#{@sym}_#{i}".intern
        else
          @sym
        end
      end

      def dynamic_element(name, opt={}, cls=DynamicElementSpec, &block)
        SpecOptionMeta::inherit_option(self, opt)
        new_spec = cls.new(name, self, opt, &block)
        @children <<  new_spec
        new_spec
      end

      
      def compile_element_main_code(element, cg, &block)
        return if @option[:dummy_element]
        @compiler.compile_main(element, self, cg, &block)
      end

      def compile_element_sub_code(element, cg, &block)
        return if @option[:dummy_element]
        @compiler.compile_sub(element, self, cg, &block)
      end

      def method_missing(*args, &block)
        name = nil
        name = args.shift if args.first.kind_of?(Symbol)
        return super unless @in_setup
        if @direct_define 
          params = args[-1] || {}
          dynamic_element(name, params, &block)
        else
          sym = nil
          sym = args.shift if args.first.kind_of?(Symbol)
          SpecOptionMeta::define_element_ext(self, name , sym, args.first, &block) or 
            super
        end
      end

      def module_name
        sym.to_s.capitalize  
      end

      def output_stream
        @option[:output_stream] or sym
      end

      def register_parts(modname=nil, method_name=nil, spec=self)
        if method_name
          if modname 
            modname = module_name + "::" + modname
          else
            modname = module_name
          end
        else
          method_name = sym.to_s 
        end
        @parent.register_parts(modname, method_name, spec)
      end

      def get_methods
        h = {}
        ret = children.find_all do |c|
          ret = h[c.sym]
          h[c.sym] = true
          ret != true and c.sym
        end.collect {|c| c.sym }
        if ret.size == 0
          children.each do |c|
            ret.concat c.get_methods
          end
        end
        ret
      end

      def make_default_template
        tag = @option[:html_tag] 
        tag = 'div' if @option[:div] 
        unless sym or tag
          ret = children.collect do |c|
            c.make_default_template
          end.join('')
          return ret
        end

        attr = @option[:html_attr]
        a = nil
        if attr
          a =  attr.collect do |k,v| 
            "#{k}='#{v}'"
          end.join(' ')
        end
        if @option[:static_data]
          tag_and_attr = tag.to_s
          tag_and_attr += ' ' + a if attr and attr.size > 0
          "<#{tag_and_attr}>#{@option[:static_data]}</#{tag}>"
        else
          tag ||= 'span' 
          tag_and_attr = tag.to_s
          tag_and_attr += ' ' + a if attr and attr.size > 0
          if sym or tag
            idparam = ""
            idparam = " #{@option[:key_attr_name] }='#{sym}'" if sym
            tag = 'span' unless tag
            if children.size > 0
              "<#{tag_and_attr}#{idparam}>" +
                children.collect do |c|
                c.make_default_template
              end.join('') + "</#{tag}>"
            else
              "<#{tag_and_attr} #{idparam} />" 
            end
          else
          end
        end
      end

      def generate_expander(mod, parent=nil)
        @compiler.generate_expander(self, mod, parent)
      end

      def get_dynamic_specs
        if sym
          [self]
        else
          children.collect do |c|
            c.get_dynamic_specs
          end.flatten
        end
      end

      def check_double_id
        h = {}
        children.each do |c|
          next unless c.sym
          raise "dobule ID #{c.sym} in #{sym || 'ROOT'}" if h[c.sym] and not c.option[:dummy_element]
          h[c.sym] = true
        end
      end

      private
      def setup_matcher
        if sym
          @matcher = @option[:matcher] || AttrMatcher.new(@option[:key_attr_name], sym.to_s)
        else
          @matcher = AnyMatcher.new
        end
      end

      def setup_compiler
        if @option[:compiler] 
          @compiler = @option[:compiler]
        else
          @compiler = DefaultCompiler.new
          SpecOptionMeta::setup_compiler(self, @compiler)
        end
      end

    end

    class RootSpec < DynamicElementSpec
      attr_reader :parts
      def initialize(option={}, &block)
        option[:key_attr_name] = :id unless option[:key_attr_name]
        super(nil, nil, option, &block)
        @parts = []
      end

      def clone
        new_spec = RootSpec.new(option)
        @children.each_with_index do |child, i|
          c = child.clone
          c.parent = new_spec
          new_spec.children[i] = c
        end
        new_spec.cloned = true
        new_spec
      end

      def module_name
        "MainContext"
      end

      def create_template(e)
        case e
        when REXML::Document
          doc = e
          root = e.root
        else
          doc = REXML::Document.new(e)
          root = doc.root
        end

        ret = Object.new
        m = create_template_module(root, doc)
        ret.extend(m)
        ret
      end

      def create_template_module(e, doc=nil)
        m = Module.new
        result = create_template_module_text(e)
        @option[:debug_source].puts result  if @option[:debug_source]
        m.module_eval result
        if doc
          m.const_set(:XmlDecl, doc.xml_decl.to_s)
          m.const_set(:DocType, doc.doctype.to_s)
        end
        m
      end

      def create_template_module_text(e)
        root = cloned ? self : self.clone
        specs = [root]
        traverse_and_search_match(specs, e)
        # not match check
        raise "SPEC #{specs.first.sym} was not matched" if specs.size > 0 

        cg = CodeGenerator.new
        cg.ie_hack = option[:ie_hack] 
        cg.code("include Amrita2::Runtime")
        cg.code("extend Amrita2::Runtime")

        cg.define_method(:expand_template, :out, "&block") do
          define_expand_template(cg, e)
        end

        cg.define_method('self.expand_template', :out, "&block") do
          define_expand_template(cg, e)
        end

        cg.define_module(module_name) do
          cg.code("Methods = #{get_methods.inspect}")
          cg.code("extend Amrita2::Runtime")
          define_sub_modules(cg, e)
        end
        root.parts.each do |modname, method_name, spec|
          cg.define_method(method_name, "*args", "&block") do
            if spec.children.size > 0
              cg.if_("block_given?") do
                cg.code('new_context() do')
                cg.level_up do
                  if modname
                    cg.code("MainContext::#{modname}::#{method_name}(*args, &block)")
                  else
                    cg.code("MainContext::#{method_name}(*args, &block)")
                  end
                  cg.code("output_substream(:#{method_name})")
                end
                cg.code("end")
                cg.else_ do 
                  cg.code("Expander.new.expand_parts(#{method_name.inspect}, self, *args)")
                end
              end
            else
              cg.code('new_context() do')
              cg.level_up do
                if modname
                  cg.code("MainContext::#{modname}::#{method_name}(*args, &block)")
                else
                  cg.code("MainContext::#{method_name}(*args, &block)")
                end
                cg.code("output_substream(:#{method_name})")
              end
              cg.code("end")
            end
          end
        end

        cg.define_method(:xml_decl) do
          cg.code("XmlDecl.dup")
        end

        cg.define_method('self.xml_decl') do
          cg.code("self::XmlDecl.dup")
        end

        cg.define_method(:doctype) do
          cg.code("DocType.dup")
        end

        cg.results.join("\n")
      end

      def define_expand_template(cg, e)
        cg.code("new_context(out) do")
        cg.level_up do
          cg.code("block.call(#{module_name or 'nil'}) if block_given? ")
          e.amrita_compile_main_code(cg)
        end
        cg.code("end")
      end

      def define_sub_modules(cg, e)
        e.amrita_compile_sub_code(cg)
      end

      def traverse_and_search_match(child_specs, element)
        spec = child_specs.first
        unless spec
          element.amrita_no_spec = true unless element.amrita_spec
          return false
        end
        if spec.matcher.match(element)
          element.amrita_no_spec = false
          element.amrita_spec = spec if spec.sym
          spec.assign_element(element)
          child_specs.shift unless spec.option[:match_many]
          specs = spec.children.collect do |s|
            s.get_dynamic_specs
          end.flatten
          #p specs,1111 if specs.size
          if spec.option[:generate_template] and specs.size > 0 and element.elements.size == 0
            generate_template(spec, element)
          end
          traverse_and_search_match(specs, element)

          # not match check
          specs.each do |s|
            next if s.option[:match_many]
            next if s.option[:static_data]
            raise "SPEC #{s.sym} was not matched" 
          end
        else
          # check children has spec and set it to amrita_no_spec and return the result
          f = false                   
          element.elements.each do |ce|
            f |= traverse_and_search_match(child_specs, ce)
          end
          element.amrita_no_spec = (not f) 
          f
        end
      end

      def generate_template(spec, element)
        src = spec.make_default_template
        #p src
        doc = REXML::Document.new src
        doc.root.elements.each do |c|
          element.add(c)
        end
      end

      def register_parts(modname, method_name, spec)
        @parts << [modname, method_name, spec]
      end

    end

    class DefaultCompiler
      def compile_main(element, dyn_spec, cg, &block)
        if dyn_spec.option[:parts_element]
          dyn_spec.register_parts
          if dyn_spec.option[:place_holder]
            cg.generate_static_element_code(element) do
            end
          end
        else
          if dyn_spec.option[:static_element]
            cg.generate_static_element_code(element) do 
              element.each_child do |c|
                c.amrita_compile_main_code(cg)
              end
            end
          else
            sym = element.amrita_spec.sym
            compile_main_code(element, dyn_spec, cg, sym.to_s, &block)
          end
        end
      end

      def compile_sub(element, dyn_spec, cg, &block)
        if dyn_spec.option[:static_element]
          element.elements.each do |c|
            c.compile_sub_code(cg)
          end
        else
          compile_dynamic_element_code(element, dyn_spec, cg, &block)
        end
      end

      def generate_expander(spec, mod, parent=nil)
        if spec.kind_of?(RootSpec)
          Expander::Root.new(spec, mod)
        else
          if spec.children.size > 0
            Expander::Node.new(parent, spec, mod)
          else
            Expander::Leaf.new(parent, spec, mod)
          end
        end
      end

      private
      def compile_main_code(element, dyn_spec, cg, sym)
        cg.code("output_substream(:#{sym})")
      end

      def compile_dynamic_element_code(element, dyn_spec, cg, &block)
        sym = dyn_spec.sym
        if dyn_spec.children.size == 0 
          code_for_leaf_element(element, dyn_spec, cg, &block)
        else
          code_for_not_leaf_element(element, dyn_spec, cg, &block)
        end
      end

      def code_for_leaf_element(element, dyn_spec, cg, &block)
        sym = dyn_spec.sym
        cg.define_element_method(sym, dyn_spec.output_stream) do
          enc = dyn_spec.option[:encoding]
          if enc
            cg.convert("utf-8", enc, "sym", "val") 
            cg.convert_hash("utf-8", enc, "attrs")
          end
          code_for_element_and_scalar(element, dyn_spec, cg)
        end
      end

      def code_for_not_leaf_element(element, dyn_spec, cg, &block)
        if dyn_spec.matched_many? and dyn_spec.first_match?(element)
          make_selection_code(element, dyn_spec, cg, &block)
        end
        sym = dyn_spec.name_for(element)
        cg.define_element_method(sym, dyn_spec.output_stream) do
          code_for_element_with_child(element, dyn_spec, cg, &block)
        end
        if ! dyn_spec.matched_many? or dyn_spec.first_match?(element)
          cg.define_module(dyn_spec.module_name) do
            cg.code("Methods = #{dyn_spec.get_methods.inspect}")
            cg.code("extend Amrita2::Runtime")
            yield
          end
        end
      end

      def make_selection_code(element, dyn_spec, cg, &block)
        sym = dyn_spec.sym
        cg.define_method(sym, "no=0", "attrs={}", "&block") do
          cg.code("no = no % #{dyn_spec.assigned_elements.size}")
          cg.code("mname = (#{sym.to_s.inspect} + '_' + no.to_s).intern")
          cg.code("self.send(mname, attrs, &block)")
        end        
        cg.code("module_function :#{sym}")
      end

      def code_for_element_and_scalar(element, dyn_spec, cg)
        cg.generate_static_element_code(element, dyn_spec.option) do 
          cg.put_expression("val.amrita_sanitize")
        end
      end

      def code_for_element_with_child(element, dyn_spec, cg)
        cg.generate_static_element_code(element, dyn_spec.option) do 
          cg.code("yield(#{dyn_spec.module_name})")
          element.each_child do |c|
            c.amrita_compile_main_code(cg)
          end
        end
      end

      def gen_element_with_attr(element,dyn_spec, cg, varname='e')
        a = element.attributes.collect do |k, v|
          next if dyn_spec.option[:delete_id] and k == dyn_spec.option[:key_attr_name].to_s
          "#{k.intern.inspect} => #{v.inspect}, "
        end
        default_value = "{ #{a.join(' ')}}"
        cg.code("a = #{default_value}")
        cg.code("attrs.each do |k,v|")
        cg.level_up do
          cg.code("a[k] = v")
        end
        cg.code("end")
        cg.code("#{varname} = new_element(#{element.name.inspect}, a)")
      end
    end
  end

  module ReplaceAttr
    extend Core::SpecOptionMeta::CompilerOption

    def code_for_element_and_scalar(element, dyn_spec, cg)
      cg.if_("attrs.size == 0") do
        cg.generate_static_element_code(element, dyn_spec.option) do 
          cg.put_expression("val.amrita_sanitize")
        end
        cg.else_ do
          gen_element_with_attr(element, dyn_spec, cg)
          cg.code("e.delete_attribute(#{dyn_spec.option[:key_attr_name].to_s.inspect})") if dyn_spec.option[:delete_id]
          cg.put_expression("start_tag(e)")
          cg.put_expression("val.amrita_sanitize")
          cg.put_expression("end_tag(e)")
        end
      end
    end

    def code_for_element_with_child(element, dyn_spec, cg)
      cg.if_("attrs.size == 0") do
        cg.generate_static_element_code(element, dyn_spec.option) do 
          cg.code("yield(#{dyn_spec.module_name}) if block_given?")
          element.each_child do |e|
            e.amrita_compile_main_code(cg)
          end
        end
        cg.else_ do
          gen_element_with_attr(element, dyn_spec, cg)
          cg.put_expression("start_tag(e)")
          cg.code("yield(#{dyn_spec.module_name})  if block_given?")
          element.each_child do |e|
            e.amrita_compile_main_code(cg)
          end
          cg.put_expression("end_tag(e)")
        end
      end
    end

    Core::SpecOptionMeta::define_option(:replace_attr, self)
  end


  def self.make_default_spec(root, type_map, option={:delete_id=>true, :check_true=>true, :replace_attr=>true, :key_attr_name=>:id})
    s = Core::RootSpec.new(option.clone)
    make_default_spec1(root, option, [s], type_map)
    s
  end

  def self.make_default_spec1(e, option, spec_stack, type_map)
    key_attr = option[:key_attr_name].to_s
    id_ = e.attributes[key_attr]
    if id_
      sym = id_.intern
      opt = get_option(e, option, type_map)
      s = Core::DynamicElementSpec.new(sym, nil, opt)
      s.parent = spec_stack.first
      spec_stack.first.children << s
      spec_stack.unshift(s)
      e.elements.each do |c|
        make_default_spec1(c, option, spec_stack, type_map)
      end
      spec_stack.shift
      spec_stack.first.check_double_id
    else
      e.elements.each do |c|
        make_default_spec1(c, option, spec_stack, type_map)
      end
    end
  end

  def self.get_option(e, option, type_map)
    ret = option.clone
    a = e.attributes
    if typ = a['amrita:type']
      default_opt = type_map[typ.intern]
      ret.merge!(default_opt) if default_opt
      a.delete('amrita:type')
    end
    if opt = a['amrita:option']
      ret.merge!(eval("{#{opt} }"))
      a.delete('amrita:option')
    end
    ret
  end

  module Expander
    class Base
      include Core::Runtime
      attr_reader :children, :name, :mod, :spec, :parent
      def initialize(parent, spec, pmod)
        @parent = parent
        @spec = spec
        @children = []
        @name = @spec.sym
        @method = pmod.method(@spec.sym || :expand_template) unless spec.option[:dummy_element]
        @parent.register_parts(self) if @parent and spec.option[:parts_element]
      end

      def register_parts(exp)
        @parent.register_parts(exp)
      end

      def expand_parts(data)
        new_context do 
          expand(data)
          output_substream(name)
        end
      end
    end

    class Leaf < Base
      def expand(data, i=nil)
        case data
        when nil
        when String, Integer, true, false, Binding
          @method.call(data)
        when REXML::Element
          if data.name
            @method.call(SanitizedString[data.to_s])
          else
            @method.call(data.text, data.attributes)
          end
        when AttrArray
          @method.call(data.body, data)
        when Enumerable
          data.each_with_index do |d, i|
            expand(d, i)
          end
        when Proc
          data.call(@parent.mod)
        else
          @method.call(data)
        end
      end
    end

    class TupleExpander < Leaf
      def expand(data, i=nil)
        case data
        when []
          super
        when Array
          if data.first.kind_of?(Array)
            super
          else
            @method.call(*data)
          end
        else
          super
        end
      end
    end

    class Node < Base
      def initialize(parent, spec, pmod)
        super
        @mod = pmod::const_get(spec.module_name)
        traverse_children(spec)
      end

      def traverse_children(spec)
        spec.children.each do |c|
          if c.sym and not c.option[:dummy_element]
            @children << c.generate_expander(@mod, self) 
          else
            traverse_children(c)
          end
        end
      end

      def each(&block)
        @children.each do |c|
          next unless c.spec.sym 
          next if c.spec.option[:dummy_element]
          yield(c)
        end
      end

      def expand(data, i=0)
        case data
        when nil
        when Proc
          data.call(@mod)
        when AttrArray
          @method.call(data, i) do 
            data = data.body
            case data
            when Hash
              each do |c|
                k = c.name
                d = data[k] || data[k.to_s]
                c.expand(d)
              end
            when Struct, Enumerable
              raise 'not implemented'
            else
              expand_po(data)
            end
          end
        when Hash
          @method.call(i) do 
            each do |c|
              k = c.name
              d = data[k] || data[k.to_s]
              c.expand(d)
            end
          end
        when Struct
          @method.call(i) do 
            expand_po(data)
          end
        when String
          raise "can not expand s String(#{data})"
        when Enumerable
          data.each_with_index do |d, i|
            expand(d, i)
          end
        else
          @method.call(i) do 
            expand_po(data)
          end
        end
      end

      def expand_po(data)
        each do |c|
          mid = c.name
          begin
            meth = data.method(mid)
            case meth.arity 
            when 0
              c.expand(meth.call)
            else
              meth.call(@mod) 
            end
          rescue NameError
            ivname = '@' + mid.to_s
            raise unless data.instance_variables.include?(ivname)
            v = data.instance_variable_get(ivname.intern)
            c.expand(v)
          end
        end
      end
    end

    class Root < Node
      def initialize(spec, pmod)
        @parts = {}
        super(nil, spec, pmod)
      end
      
      def register_parts(exp)
        @parts[exp.name] = exp
      end

      def get_parts(name)
        @parts[name]
      end

      def expand_template(out,data)
        case data
        when Proc
          data.call(@mod)
        when Hash
          @method.call(out) do 
            @children.each do |c|
              k = c.name
              d = (data[k] || data[k.to_s])
              c.expand(d)
            end
          end
        else
          @method.call(out) do 
            expand_po(data)
          end
        end
      end
    end
  end


  module HtmlTag
    include Amrita2::Core
    HEAD = TagMatcher.new(:head)
    LINK = TagMatcher.new(:link)
    TITLE = TagMatcher.new(:title)
    BODY = TagMatcher.new(:body)
    TABLE = TagMatcher.new(:table)
    TR = TagMatcher.new(:tr)
    TH = TagMatcher.new(:th)
    TD = TagMatcher.new(:td)
    UL = TagMatcher.new(:ul)
    OL = TagMatcher.new(:ol)
    LI = TagMatcher.new(:li)
    DL = TagMatcher.new(:dl)
    DT = TagMatcher.new(:dt)
    DD = TagMatcher.new(:dd)
    A = TagMatcher.new(:a)
    PRE = TagMatcher.new(:pre)
    DIV = TagMatcher.new(:div)
    SPAN = TagMatcher.new(:span)
    P = TagMatcher.new(:p)
    FORM = TagMatcher.new(:form)
  end
  class Expander_old
    def initialize(opt={})
      @option = opt
    end

    def expand(t, out, data)
      t.expand_template(out) do |m|
        expand1(m, data)
      end
    end

    def expand_parts(parts_mid, t, *data)
      t.send(parts_mid) do |m|
        expand1(m, *data)
      end
    end

    def expand1(m, data)
      if data.kind_of?(Proc)
        data.call(m)
      else
        methods = m::Methods
        if data.kind_of?(Hash)
          expand_hash(m, data, methods)
        else
          expand_po(m, data, methods)
        end
      end
    end

    def expand_hash(m, h, methods)
      methods.each do |k|
        v = h[k] || h[k.to_s]
        expand_data(m, k, v)
      end
    end

    #     def expand_po(m, po, methods)
    #       methods.each do |mid|
    #         expand_data(m, mid, po.send(mid))
    #       end
    #     end

    def expand_po(m, po, methods)
      methods.each do |mid|
        meth = po.method(mid)
        case meth.arity 
        when 0
          expand_data(m, mid, meth.call)
        else
          meth.call(m) 
        end
      end
    end

    def expand_data(m, k, v, no=nil)
      case v
      when Hash, Struct
        m.send(k, no) do |m2|
          expand1(m2, v)
        end
      when String, Integer, true, false, nil, Binding
        m.send(k, v) do |m2|
        end
      when REXML::Element
        if v.name
          m.send(k, SanitizedString[v.to_s]) do |m2|
            expand(m2, v.text) if t.text
          end
        else
          m.send(k, v.text, v.attributes) do |m2|
          end
        end
      when Enumerable
        v.each_with_index do |vv, i|
          expand_data(m, k, vv, i)
        end
      when Proc
        v.call(m)
      else
        m.send(k, no) do |m2|
          expand1(m2, v)
        end
      end
    end

  end

#   class ExpanderForActivePO < Expander
#     def expand_po(m, po, methods)
#       methods.each do |mid|
#         begin
#           meth = po.method(mid)
#           case meth.arity 
#           when 0
#             expand_data(m, mid, meth.call)
#           else
#             meth.call(m) 
#           end
#         rescue NameError
#           ivname = '@' + mid.to_s
#           v = po.instance_variable_get(ivname.intern)
#           expand_data(m, mid, v)
#         end
#       end
#     end
#   end

  Runtime = Core::Runtime
  def self.define_template_spec(option={}, &block)
    Core::RootSpec.new(option, &block)
  end

  class AttrArray < Hash
    attr_reader :body
    def initialize(h, &block)
      h.each do |k, v|
        self[k] = v
      end
      if block_given?
        @body = yield
      else
        @body = nil
      end
    end
  end

  module REXMLShortCut

    def e(tag, hash={})
      ret = REXML::Element.new(tag.to_s)
      hash.each do |k, v|
        ret.attributes[k.to_s] = v
      end
      if block_given?
        case child = yield 
        when REXML::Element
          ret.add child
        else
          ret.text = child
        end
      end
      ret
    end
    module_function :e

    def a(hash={}, &block)
      AttrArray.new(hash, &block)
    end
    module_function :a
  end
  include REXMLShortCut
end

