
require 'test/unit'
require 'test/unit/testsuite'
require 'test/unit/ui/console/testrunner'
require 'amrita2/a2tdl'
require 'rexml/document'
require 'date'
require 'open3'

# ruby -I~/amrita2/lib ~/amrita2/test/test_a2tdl.rb   |ruby -pe '$_.sub!(/^ */, "") ' 

module Amrita2GeneratorTest
  include Amrita2
  include Amrita2::Core
  include Amrita2::TextSupport


  class TestForm  < Test::Unit::TestCase
    def test_input
      spec = Amrita2::define_template_spec do 
        form_element(:aaa) do 
          text_field(:bbb)
        end
      end
      doc = REXML::Document.new  "<form id='aaa'>111<input name='bbb' />222</form>"
      t = spec.create_template(doc)

      root = doc.root
      assert(root.amrita_spec)
      assert(:aaa, root.amrita_spec.sym)

      out = ""
      t.expand_template(out) do |m|
        m.aaa do |mm|
          mm.bbb('abc')
        end
      end

      assert_equal("<form id='aaa'>111<input name='bbb' type='text' value='abc'/>222</form>", out)
    end

    def test_input2
      spec = Amrita2::define_template_spec do 
        form_element(:aaa, :generate_template=>true) do 
          text_field(:bbb)
        end
      end
      doc = REXML::Document.new  "<form id='aaa' />"
      t = spec.create_template(doc)

      out = ""
      t.expand_template(out) do |m|
        m.aaa do |mm|
          mm.bbb('abc')
        end
      end

      assert_equal("<form id='aaa'><input name='bbb' type='text' value='abc'/></form>", out)
    end

    def test_textarea
      spec = Amrita2::define_template_spec do 
        form_element(:aaa, :generate_template=>true) do 
          textarea(:bbb, :html_attr=>{ :col=>123 })
        end
      end
      doc = REXML::Document.new  "<form id='aaa' />"
      t = spec.create_template(doc)
      out = ""
      t.expand_template(out) do |m|
        m.aaa do |mm|
          mm.bbb('abc')
        end
      end

      assert_equal("<form id='aaa'><textarea col='123' name='bbb'>abc</textarea></form>", out)
    end

    def test_select
      spec = Amrita2::define_template_spec do 
        form_element(:form1) do 
          select(:sel1) do |s|
            s.opt(:a, 'aaa')
            s.opt(:b, 'bbb')
            s.opt(:c, 'ccc')
          end
        end
      end
      doc = REXML::Document.new  "<form id='form1'><select name='sel1' /></form>"
      t = spec.create_template(doc)

      root = doc.root
      assert(root.amrita_spec)
      assert(:aaa, root.amrita_spec.sym)

      out = ""
      t.expand_template(out) do |m|
        m.form1 do |mf|
          mf.sel1(:a)
          mf.sel1(:b)
        end
      end
      assert_equal("<form id='form1'><select name='sel1'><option selected='selected' value='a'>aaa</option><option value='b'>bbb</option><option value='c'>ccc</option></select><select name='sel1'><option value='a'>aaa</option><option selected='selected' value='b'>bbb</option><option value='c'>ccc</option></select></form>", out)

      doc = REXML::Document.new  "<form id='form1'><select name='sel1'><option value='a'>xxx</option><option value='b'>yyy</option><option value='c'>zzz</option></select></form>"
      t = spec.create_template(doc)

      root = doc.root
      assert(root.amrita_spec)
      assert(:aaa, root.amrita_spec.sym)

      out = ""
      t.expand_template(out) do |m|
        m.form1 do |mf|
          mf.sel1(:c)
        end
      end
      assert_equal("<form id='form1'><select name='sel1'><option value='a'>xxx</option><option value='b'>yyy</option><option selected='selected' value='c'>zzz</option></select></form>", out)
    end
    
    def test_radio
      spec = Amrita2::define_template_spec do 
        form_element(:form1) do 
          radio_buttons(:radio) do |s|
            s.opt(:a, 'aaa')
            s.opt(:b, 'bbb')
            s.opt(:c, 'ccc')
          end
        end
      end
      doc = REXML::Document.new  "<form id='form1'><fieldset name='radio' /></form>"
      t = spec.create_template(doc)

      root = doc.root
      assert(root.amrita_spec)
      assert(:aaa, root.amrita_spec.sym)

      out = ""
      t.expand_template(out) do |m|
        m.form1 do |mf|
          mf.radio(:c)
        end
      end
      assert_equal("<form id='form1'><fieldset name='radio'><input name='radio' type='radio' value='a'>aaa</input><input name='radio' type='radio' value='b'>bbb</input><input checked='checked' name='radio' type='radio' value='c'>ccc</input></fieldset></form>",out)

    end

    def test_grid_form
      t = Template.new do
        Amrita2::define_template_spec do 
          grid_form(:f1, :generate_template=>true) do 
            text_field(:name, :description=>'input your name')
            password_field(:pass, :description=>'input password')
            text_field(:email)
            check_box(:amrita2_user)
          end
        end
      end

      out = ""
      data = { 
        :name => 'abc',
        :pass => '',
        :email => 'tnakajima@brain-tokyo.jp',
        :amrita2_user => :checked
      }
      t.expand(out, :f1=>data)
      #print out
      #assert_equal("<form id='f1'>\n<table>\n  <tr><th>input your name</th><td><input name='name' type='text' value='abc'/></td></tr><tr><th>input password</th><td><input name='pass' type='password' value=''/></td></tr><tr><th>email</th><td><input name='email' type='text' value='tnakajima@brain-tokyo.jp'/></td></tr><tr><th>amrita2_user</th><td><input name='amrita2_user' type='checkbox'/></td></tr>\n</table>\n</form>\n", out)
    end

  end

  class TestHtml  < Test::Unit::TestCase
    include Amrita2::Runtime
    def assert_equal_xml(a, b)
      a = REXML::Document.new(a, :ignore_whitespace_nodes=>:all).to_s
      b = REXML::Document.new(b, :ignore_whitespace_nodes=>:all).to_s
      assert_equal(a, b)
    rescue 
      puts a
      puts 
      puts b
      raise
    end

    def test_html_gen
      t = Amrita2::Template.new do
        Amrita2::define_template_spec do 
          div do
            span(:test) 
          end
        end
      end
      t.expand(out="") do |m|
        m.test('testtest')
      end
      assert_equal("<div><span id='test'>testtest</span></div>", out)
    end

    def test_html_gen2
      spec = Amrita2::define_template_spec do 
        html do
          hr
          h1('aaaa')
          span(:test) 
          p('xxxx')
        end
      end
      #p spec
      t = Amrita2::Template.new { spec.clone }
      t.expand(out="") do |m|
        m.test('testtest')
      end
      assert_equal_xml("<html><body><hr></hr><h1>aaaa</h1><span id='test'>testtest</span><p>xxxx</p></body></html>", out)

      t = Amrita2::TemplateText.new("<html><hr /><span id='test'/><hr /></html>") { spec.clone }
      t.expand(out="") do |m|
        m.test('test2')
      end
      assert_equal("<html><hr/><span id='test'>test2</span><hr/></html>", out)
    end

    def test_html_gen3
      spec = Amrita2::define_template_spec do 
        html do
          div(:abc, :delete_id=>true) 
        end
      end
      #p spec
      t = Amrita2::Template.new { spec.clone }
      t.expand(out="", :abc=>'111') 
      assert_equal_xml("<html><body><div>111</div></body></html>", out)
    end

    def test_html_gen4
      spec = Amrita2::define_template_spec do 
        html do
          div(:navigation) do 
            a_link(:prev, :html_attr=>{ :align => 'left'})
            a_link(:next, :html_attr=>{ :align => 'right'})
          end
          hr
          h1('aaaa')
          div(:group, :delete_id=>true) do
            h2(:group_title)
            span(:aaa) 
            p('xxxx')
          end
        end
      end
      #p spec
      t = Amrita2::Template.new { spec.clone }
      t.expand(out="") do |m|
        m.group do |mg|
          mg.group_title('g1')
          mg.aaa('1')
        end
        m.group do |mg|
          mg.group_title('g2')
          mg.aaa('2')
        end
      end
      assert_equal_xml("<html><body><hr/><h1>aaaa</h1><div><h2>g1</h2><span>1</span><p>xxxx</p></div><div><h2>g2</h2><span>2</span><p>xxxx</p></div></body></html>", out)

    end

    def test_html_title
      t = Amrita2::Template.new do
        Amrita2::define_template_spec do 
          html(:css=>'a.css') do
            title('tttt')
            span(:test) 
          end
        end
      end
      t.expand(out="") do |m|
        m.test('testtitle')
      end
      assert_equal_xml("<html><head><title>tttt</title><link href='a.css' rel='stylesheet' type='text/css'></link></head><body><h1>tttt</h1><span id='test'>testtitle</span></body></html>", out)
    end

    def xx_test_table
      t = TemplateText.new "<div><table>\n  <tr><th>aaa</th><th>bbb</th></tr>\n  <tr id='item'><td id='aaa'/><td id='bbb'/></tr><tr id='item' class='b'><td id='aaa'/><td id='bbb'/></tr>\n</table>\n</div>" do
        Amrita2::define_template_spec do
          dynamic_element(nil) do
            dynamic_element(:item, { :match_many=>true, :table_opt=>{:name=>:item}, :key_attr_name=>:id}) do
              dynamic_element(:aaa, {:key_attr_name=>:id})
              dynamic_element(:bbb, {:key_attr_name=>:id})
            end
          end
        end
      end
      data = [
        { :aaa=>1, :bbb=>2},
        { :aaa=>3, :bbb=>4},
      ]
      t.expand(out="", :item=>data) 
      #print out
    end

    def test_html_table
      t = Amrita2::Template.new do
        Amrita2::define_template_spec do 
          div do 
            simple_table(:item) do 
              dynamic_element(:aaa)
              dynamic_element(:bbb)
            end
          end
        end
      end
      #t.setup; p t.spec
      data = [
        { :aaa=>1, :bbb=>2},
        { :aaa=>3, :bbb=>4},
      ]
      t.expand(out="", :item=>data) 
      #puts out

      assert_equal_xml("<div><table><tr><th>aaa</th><th>bbb</th></tr><tr><td>1</td><td>2</td></tr><tr><td>3</td><td>4</td></tr></table></div>",out)
    end

    def test_html_table2
      t = Amrita2::Template.new do
        Amrita2::define_template_spec do 
          div do 
            simple_table(:item, :table_opt=>{ :classes=>%w(odd even), :border=>1}) do 
              column(:aaa, :table_opt=>{ :align=>'center'})
              column(:bbb)
            end
          end
        end
      end
      #t.setup; p t.spec
      data = [
        { :aaa=>1, :bbb=>2},
        { :aaa=>3, :bbb=>4},
        { :aaa=>5, :bbb=>6},
        { :aaa=>7, :bbb=>8},
      ]
      t.expand(out="") do |m|
        data.each_with_index do |item, i|
          m.item(i) do |mi|
            mi.aaa(item[:aaa])
            mi.bbb(item[:bbb])
          end
        end
      end
      #puts out

      assert_equal_xml(<<-END, out)
<div><table border='1'>
  <tr><th>aaa</th><th>bbb</th></tr>
  <tr class='odd'> <td align='center'>1</td><td>2</td></tr>
  <tr class='even'><td align='center'>3</td><td>4</td></tr>
  <tr class='odd'> <td align='center'>5</td><td>6</td></tr> 
  <tr class='even'><td align='center'>7</td><td>8</td></tr>
</table></div>
END
      t.expand(out2="", :item=>data)
      assert_equal_xml(out, out2)
    end

    def test_partial
      spec = Amrita2::define_template_spec(:generate_template=>true) do 
        dynamic_element(:a) do 
            div do 
              p('xxxx')
              span(:test) 
            end
        end
      end
      t = Amrita2::TemplateText.new "<xxx id='a'></xxx>"
      t.spec = spec 
      t.expand(out="") do |m|
        m.a do |m2|
          m2.test('123')
        end
      end
      assert_equal("<xxx id='a'><div><p>xxxx</p><span id='test'>123</span></div></xxx>", out)
      t = Amrita2::TemplateText.new "<xxx id='a'>yyy<span id='test'>123</span>xxx</xxx>"
      t.spec = spec 
      t.expand(out="") do |m|
        m.a do |m2|
          m2.test('123')
        end
      end
      assert_equal("<xxx id='a'>yyy<span id='test'>123</span>xxx</xxx>",out)
    end

    def test_partial2
      spec = Amrita2::define_template_spec(:generate_template=>true) do 
        dynamic_element(:a) do 
          dynamic_element(:b) do 
            div do 
              p('xxxx')
              span(:test) 
            end
          end
        end
      end

      t = Amrita2::TemplateText.new "<yyy id='a'><xxx id='b'  /></yyy>"
      t.spec = spec 
      t.expand(out="") do |m|
        m.a do |m2|
          m2.b do |m3|
            m3.test('123')
          end
        end
      end
      assert_equal("<yyy id='a'><xxx id='b'><div><p>xxxx</p><span id='test'>123</span></div></xxx></yyy>", out)
      t = Amrita2::TemplateText.new "<xxx id='a'><yyy id='b'><span id='test'>123</span>xxx</yyy></xxx>"
      t.spec = spec 
      t.expand(out="") do |m|
        m.a do |m2|
          m2.b do |m3|
            m3.test('123')
          end
        end
      end
      assert_equal("<xxx id='a'><yyy id='b'><span id='test'>123</span>xxx</yyy></xxx>",out)
    end

    def test_erb
      t = Amrita2::Template.new do
        Amrita2::define_template_spec do 
          div do
            erb(:nlist, <<-END, :use_single_arg=>true)
            <ul>
            <% $_.times do |i| %>
              <li><%= i %></li>
            <% end %>
            </ul>
            END
          end
        end
      end
      t.expand(out="") do |m|
        m.nlist(3)
      end
      assert_equal_xml("<div><ul><li>0</li><li>1</li><li>2</li></ul></div>", out)
    end

    def test_erb_args
      t = Amrita2::Template.new do
        Amrita2::define_template_spec do 
          div do
            erb(:plus, "<%=$1 + $2 %>", :use_erb_args=>true)
          end
        end
      end
      t.expand(out="") do |m|
        m.plus(1, 2)
      end
      assert_equal("<div>3</div>", out)
    end

    def test_erb_args2
      t = Amrita2::Template.new do
        Amrita2::define_template_spec do 
          div do
            erb(:link, <<-END, :use_erb_args=>true)
            <a href="<%= $1 %>"><%=$0%></a>
            END
          end
        end
      end
      t.expand(out="") do |m|
        m.link('http://amrita2.rubyforge.org/') { 'Amrita2 Home Page' }
      end
      assert_equal_xml("<div><a href='http://amrita2.rubyforge.org/'>Amrita2 Home Page</a></div>", out)
    end

    def erb_binding_test(a)
      'erb' + a
    end

    def test_erb_binding
      t = Amrita2::Template.new do
        Amrita2::define_template_spec do 
          div do
            erb(:erb_here, '<%= erb_binding_test x %>')
          end
        end
      end
      t.expand(out="") do |m|
        x = 'xxx11'
        m.erb_here(binding)
      end
      assert_equal("<div>erbxxx11</div>", out)
      x = 'xxx22'
      t.expand(out="", :erb_here=>binding)
      assert_equal("<div>erbxxx22</div>", out)
    end

    def test_link
      t = Amrita2::Template.new do
        Amrita2::define_template_spec do 
          div do
            a_link(:test, :delete_id=>true) 
          end
        end
      end
      t.expand(out="") do |m|
        m.test('http://test/', 'testtest')
      end
      assert_equal("<div><a href='http://test/'>testtest</a></div>", out)

      t.expand(out="") do |m|
        m.test('javascript:allert("hello") ;', 'testtest')
      end
      assert_equal("<div><a href=''>testtest</a></div>", out)
    end

    def test_link2
      t = Amrita2::TemplateText.new("<a id='test' class='x'/>") do
        Amrita2::define_template_spec do 
          div do
            a_link(:test, :delete_id=>true) 
          end
        end
      end
      t.expand(out="") do |m|
        m.test('http://test/', 'testtest')
      end
      assert_equal("<a href='http://test/' class='x'>testtest</a>", out)
      t.expand(out="") do |m|
        m.test('http://test/', 'test', :class=>'y')
      end
      assert_equal("<a href='http://test/' class='y'>test</a>", out)
    end

    def test_form_in_html
      t = Template.new do
        Amrita2::define_template_spec do 
          html do 
            form_element(:form) do 
              check_box(:item_done)
            end
          end
        end
      end
      t.expand(out="", :form=>{ :item_done=>true })
      assert_equal_xml("<html><body><form id='form'><input checked='true' name='item_done' type='checkbox'/></form></body></html>", out)
    end

    def test_form_in_html2
      t = Template.new do
        Amrita2::define_template_spec do 
          html do 
            form_element(:form, :html_attr=>{ :method=>'post', :action=>'add_item' }) do 
              span('New item:')
              text_field(:description , :html_attr=>{ :size=>'30', :name=>'new_item[description]'})
              submit(:ok, :html_attr=>{ :value=>'Add item'})
            end
          end
        end
      end
      #t.setup; p t.spec
      t.expand(out="", :form=> { :description=>"", :ok=>true})
      #puts out
      assert_equal_xml("<html><body><form action='add_item' method='post' id='form'><span>New item:</span><input name='description' size='30' type='text' value=''/><input name='ok' type='submit' value='true'/></form></body></html>", out)
    end

    def test_dynamic_in_static
      t = Template.new do
        Amrita2::define_template_spec do 
          div do 
            ul do
              li(:list, :delete_id=>true)
            end
          end
        end
      end
      #t.setup; p t.spec
      t.expand(out="", :list=>%w(1 2 3))
      #puts out
      assert_equal_xml("<div><ul><li>1</li><li>2</li><li>3</li></ul></div>", out)
    end

    def test_callender
      t = Amrita2::Template.new do
        Amrita2::define_template_spec do 
          simple_calender(:c1) do 
            cal_date do
              a_link(:sched)
            end
          end
        end
      end

      data = { 
        :c1 => TDL::Calender::Data.new(2005,4) do |d|
          if d.day % 4 == 0
            { :sched=> a(:href=>"?#{d.day}") { "*#{ d.day}" }}
          else
            nil
          end
        end
      }
      t.expand(out="", data)
      #puts out
      formated_out = Open3.popen3("w3m -T text/html") do |i, o, e|
        i.puts out
        i.close
        o.read
      end
      #puts formated_out
      assert_equal(<<-END, formated_out)
          2005/4
Sun Mon Tue Wed Thu Fri Sat
                      1   2
  3  *4   5   6   7  *8   9
 10  11 *12  13  14  15 *16
 17  18  19 *20  21  22  23
*24  25  26  27 *28  29  30

END
    end

    def test_callender2
      t = Amrita2::Template.new do
        Amrita2::define_template_spec do 
          div do
          simple_calender(:c1) do 
            cal_date do
              span(:d)
              div (:sched) do
                a_link(:link)
              end
            end
          end
        end
      end
      end

      schedule = { 
        Date.new(2005,4,10) => { 
          :link=>a(:href=>'http://amrita2.rubyforge.org/') {
            'sch1' 
          }
        },
        Date.new(2005,4,15) => { 
          :link => a(:href=>'http://amrita2.rubyforge.org/') {
            'sch2' 
          }
        },
        Date.new(2005,4,25) => [
          { 
            :link => a(:href=>'http://amrita2.rubyforge.org/') {
              'sch31' 
            }
          },
          { 
            :link => a(:href=>'http://amrita2.rubyforge.org/') {
              'sch32' 
            }
          },
        ]
      }

      data = { 
        :c1 => TDL::Calender::Data.new(2005,4) do |d|
          sch = schedule[d]
          if sch
            { :sched =>  sch, :d=>d.day } 
          end
        end
      }
      t.expand(out="", data)
      #puts out
      formated_out = Open3.popen3("w3m -T text/html") do |i, o, e|
        i.puts out
        i.close
        o.read
      end
      #puts formated_out
      assert_equal(<<END, formated_out)
             2005/4
Sun   Mon   Tue Wed Thu Fri  Sat
                           1   2
   3      4   5   6   7    8   9
  10     11  12  13  14   15  16
sch1                    sch2
  17     18  19  20  21   22  23
         25
  24  sch31  26  27  28   29  30
      sch32

END
    end


    def test_hello_world
      t = Template.new do
        Amrita2::define_template_spec do 
          html do 
            title('Hello world')
            h2(:title)
            p(:body)
            h2('Static part')
            p('String is static part. Symbol is dynamic part')
            ul do    
              li(:contents, :delete_id=>true) do 
                a_link(:link)
              end
            end
          end
        end
      end
      #t.setup; p t.spec
      data = { 
        :title => 'Hello ROR with Amrita2',
        :body => 'Using Amrita2 with ROR is easy and fun!',
        :contents => [
          ['index', 'hello world '],
          ['index_tdl', 'TDL version hello world '],
          ['languages', 'languages: making talbe with po_method '],
          ['languages_tdl', 'TDL version languages'],
        ].collect do |url, text|
          { :link=>a(:href=>url) { text } }
        end
      }
      t.expand(out="", data)
      #puts out
      assert_equal_xml("<html><head><title>Hello world</title></head><body><h1>Hello world</h1><h2 id='title'>Hello ROR with Amrita2</h2><p id='body'>Using Amrita2 with ROR is easy and fun!</p><h2>Static part</h2><p>String is static part. Symbol is dynamic part</p><ul><li><a href='index'>hello world </a></li><li><a href='index_tdl'>TDL version hello world </a></li><li><a href='languages'>languages: making talbe with po_method </a></li><li><a href='languages_tdl'>TDL version languages</a></li></ul></body></html>", out)
    end
  end
end

if __FILE__ == $0
  include Amrita2GeneratorTest
  cname = ARGV.shift
  if cname
    method = ARGV.shift
    if method
      eval "Test::Unit::UI::Console::TestRunner.run(#{cname}.new(\"#{method}\"))"
    else
      cls = eval(cname)
      Test::Unit::UI::Console::TestRunner.run(cls)
    end
  else
    suite = Test::Unit::TestSuite.new
    suite << TestForm.suite
    suite << TestHtml.suite
    Test::Unit::UI::Console::TestRunner.run(suite)
  end
end
