

require 'test/unit'
require 'test/unit/testsuite'
require 'test/unit/ui/console/testrunner'
require 'amrita2/template'
require 'amrita2/ext'
require 'rexml/document'
require 'open3'

# ruby -I~/amrita2/lib ~/amrita2/test/test_template.rb   |ruby -pe '$_.sub!(/^ */, "") ' 

module Amrita2TemplateTest

  class TestDefaultSpec < Test::Unit::TestCase
    include Amrita2
    
    def test_default_spec
      doc = REXML::Document.new "<a id='aaa'><b id='bbb'></b></a>"
      root = doc.root
      spec = Amrita2::make_default_spec(root, Template::DefaultTypeMap)
      assert_equal(nil, spec.sym)
      assert_equal(1, spec.children.size)
      assert_equal(:aaa, spec.children[0].sym)
      assert_equal(1, spec.children[0].children.size)
      assert_equal(:bbb, spec.children[0].children[0].sym)
      assert_equal(0, spec.children[0].children[0].children.size)

      t = spec.create_template(doc)

      root = doc.root
      assert(root.amrita_spec)
      assert(:aaa, root.amrita_spec.sym)
      assert(root.elements["b"].amrita_spec)
      assert(:aaa, root.elements["b"].amrita_spec.sym)

      out = ""
      t.expand_template(out) do |m|
        m.aaa do |m2|
          m2.bbb('yyy')
        end
      end

      assert_equal("<a><b>yyy</b></a>", out)
    end

    def test_default_spec2
      doc = REXML::Document.new <<-END
<html>
   <body>
      <h1 id='title'>title will be inserted here</h1>
      <p id='body'>body text will be inserted here</p>
   </body>
</html>
END

      root = doc.root
      spec = Amrita2::make_default_spec(root, Template::DefaultTypeMap)
      assert_equal(nil, spec.sym)
      assert_equal(2, spec.children.size)
      assert_equal(:title, spec.children[0].sym)
      assert_equal(0, spec.children[0].children.size)
      assert_equal(:body, spec.children[1].sym)

      t = spec.create_template(doc)

      root = doc.root
      assert(root.elements["body/h1"].amrita_spec)
      assert(:aaa, root.elements["body/h1"].amrita_spec.sym)

      out = ""
      t.expand_template(out) do |m|
        m.title('ttt')
        m.body('bbb')
      end

      assert_equal("<html>\n   <body>\n      <h1>ttt</h1>\n      <p>bbb</p>\n   </body>\n</html>", out)
    end
    def test_error
      assert_raise(RuntimeError) do
        doc = REXML::Document.new "<a id='aaa'><b id='bbb'></b><c id='bbb' /></a>"
        root = doc.root
        spec = Amrita2::make_default_spec(root, Template::DefaultTypeMap)
      end
    end
  end

  class TestExpander < Test::Unit::TestCase
    include Amrita2
    class A
      def aaa
        B.new
      end

      def struct_array
        [B.new, B.new]
      end
    end

    class B
      def bbb
        'xxx'
      end
      
      def array
        [1,2,3]
      end

      def hash
        {:a=>'b', :c=>'d'}
      end

      def hasharray
        [ hash, hash]
      end
    end

    def test_expander
      t = TemplateText.new "<a id='aaa'><b id='bbb'></b></a>" 
      t.expand(out="", A.new) 
      assert_equal("<a><b>xxx</b></a>", out)
    end

    def test_array
      t = TemplateText.new "<a id='aaa'><b id='array'></b></a>" 
      t.expand(out="", A.new) 
      assert_equal("<a><b>1</b><b>2</b><b>3</b></a>", out)
    end

    def test_struct_array
      t = TemplateText.new "<a id='struct_array'><b id='bbb'></b></a>" 
      t.expand(out="", A.new) 
      assert_equal("<a><b>xxx</b></a><a><b>xxx</b></a>", out)
    end


    def test_hash
      t = TemplateText.new "<a id='aaa'><b id='hash'><c id='a'></c></b></a>" 
      t.expand(out="", A.new) 
      assert_equal("<a><b><c>b</c></b></a>", out)
    end

    def test_hasharray
      t = TemplateText.new "<a id='aaa'><b id='hasharray'><c id='a'></c></b></a>" 
      t.expand(out="", A.new) 
      assert_equal("<a><b><c>b</c></b><b><c>b</c></b></a>", out)
    end

    class IVHolder
      def initialize(a, b)
        @aaa, @bbb = a, b
      end
    end

    def test_iv
      t = TemplateText.new "<div><a id='aaa'></a><b id='bbb'></b></div>" 
      t.expand(out="", IVHolder.new('xxx', 'yyy'))
      assert_equal("<div><a>xxx</a><b>yyy</b></div>", out)
    end

    def test_attributes
      t = TemplateText.new "<div><a id='aaa'></a></div>" 
      t.expand(out="", :aaa=>a('href'=>'http://rexml/') { 'test' })
      assert_equal("<div><a href='http://rexml/'>test</a></div>", out)
    end

    def test_attributes2
      t = TemplateText.new "<div><a id='aaa'><em id='bbb' /></a></div>" 
      t.expand(out="", :aaa=>a('href'=>'http://rexml/') { { :bbb=>'test'} })
      assert_equal("<div><a href='http://rexml/'><em>test</em></a></div>", out)
    end

    def test_attributes3
      t = TemplateText.new "<div><a id='aaa'><span id='year' />/<span id='month' /></a></div>" 
      t.expand(out="", :aaa=>a('href'=>'http://rexml/') { Time.at(0) })
      assert_equal("<div><a href='http://rexml/'><span>1970</span>/<span>1</span></a></div>", out)
    end

    def test_element
      t = TemplateText.new "<div><span id='aaa'></span></div>" 
      t.expand(out="", :aaa=>e(:a, 'href'=>'http://rexml/') { 'test' })
      assert_equal("<div><span><a href='http://rexml/'>test</a></span></div>", out)
      t.expand(out="", :aaa=>e(:a, 'href'=>'http://rexml/') { e(:em){ 'test'} })
      assert_equal("<div><span><a href='http://rexml/'><em>test</em></a></span></div>", out)
    end

  end

#   class TestDefaultExpand < Test::Unit::TestCase
#     include Amrita2

#     def test_default_spec
#       doc = REXML::Document.new "<a id='aaa'><b id='bbb'></b></a>"
#       root = doc.root
#       spec = Amrita2::make_default_spec(root)

#       t = spec.create_template(doc)

#       root = doc.root
#       assert(root.amrita_spec)
#       assert(:aaa, root.amrita_spec.sym)
#       assert(root.elements["b"].amrita_spec)
#       assert(:aaa, root.elements["b"].amrita_spec.sym)

#       data = {
#         :aaa =>  {
#           :bbb => "yyy"
#         }
#       }
#       data = {
#         'aaa' =>  {
#           'bbb' => "yyy"
#         }
#       }
#       out = ""
#       Expander.new.expand(t, out, data)
#       assert_equal("<a><b>yyy</b></a>", out)

#       data = {
#         :aaa =>  {
#           :bbb => [1,2,3]
#         }
#       }
#       out = ""
#       Expander.new.expand(t, out, data)
#       assert_equal("<a><b>1</b><b>2</b><b>3</b></a>", out)
#     end
#   end

  class TestTemplate < Test::Unit::TestCase
    include Amrita2

    def test_template
      t = TemplateText.new "<a id='aaa'><b id='bbb'></b></a>" 
      t.expand(out="") do |m|
        m.aaa do |m2|
          m2.bbb('xxx')
        end
      end
      assert_equal("<a><b>xxx</b></a>", out)
    end

    def test_spec
      text = "<a id='aaa'><b id='bbb'></b></a>" 
      t = TemplateText.new(text) 
      t.spec =Amrita2::define_template_spec do
        dynamic_element(:aaa) do 
          dynamic_element(:bbb, :use_original_element=>true)
        end
      end
      t.expand(out="") do |m|
        m.aaa do |m2|
          m2.bbb do |e|
            e.attributes['class'] = 'using_spec'
            e.text = 'using spec'
            e
          end
        end
      end
      assert_equal("<a id='aaa'><b class='using_spec' id='bbb'>using spec</b></a>", out)
    end

    def test_spec2
      text = "<a id='aaa'><b id='bbb'></b></a>" 
      t = TemplateText.new(text)  do
        Amrita2::define_template_spec do
          dynamic_element(:aaa) do 
            dynamic_element(:bbb, :use_original_element=>true)
          end
        end
      end
      t.expand(out="") do |m|
        m.aaa do |m2|
          m2.bbb do |e|
            e.attributes['class'] = 'using_spec2'
            e.text = 'using spec2'
            e
          end
        end
      end
      assert_equal("<a id='aaa'><b class='using_spec2' id='bbb'>using spec2</b></a>", out)
    end

    def test_spec3
      text = "<a id='aaa'><b id='bbb'></b></a>" 
      t = TemplateText.new(text)
      def t.get_spec
        Amrita2::define_template_spec do
          dynamic_element(:aaa) do 
            dynamic_element(:bbb, :use_original_element=>true)
          end
        end
      end
      t.expand(out="") do |m|
        m.aaa do |m2|
          m2.bbb do |e|
            e.attributes['class'] = 'using_spec3'
            e.text = 'using spec3'
            e
          end
        end
      end
      assert_equal("<a id='aaa'><b class='using_spec3' id='bbb'>using spec3</b></a>", out)
    end

    def test_spec4
      text = "<x id='aaa'><y id='bbb'>1</y><y id='bbb'>2</y></x>"
      t = TemplateText.new(text) do
        Amrita2::define_template_spec do
          dynamic_element(:aaa) do
            dynamic_element(:bbb) 
            dummy_element(:bbb) # 222 
          end
        end
      end

      data = {"aaa"=>{:bbb=>["111", "222"]}}
      t.expand(out="", data)
      assert_equal("<x id='aaa'><y id='bbb'>111</y><y id='bbb'>222</y></x>", out)
    end

    class TestPO
      class Child
        def bbb
          'xxx'
        end
      end
      def aaa
        Child.new
      end

      def active(m)
        m.active do |m|
          m.bbb('xxx')
          m.bbb('yyy')
          m.bbb('zzz')
        end
      end

      def attr(m)
        m.attr(:aaa=>'bbb') do |m|
          m.bbb('xxx', :yyy=>'zzz')
        end
      end
    end

    def test_presentatio_object
      t = TemplateText.new "<a id='aaa'><b id='bbb'></b></a>" 
      t.expand(out="", TestPO.new) 
      assert_equal("<a><b>xxx</b></a>", out)

      t = TemplateText.new "<a id='active'><b id='bbb'></b></a>" 
      t.expand(out="", TestPO.new) 
      assert_equal("<a><b>xxx</b><b>yyy</b><b>zzz</b></a>", out)


      t = TemplateText.new "<x id='outer'><a id='active'><b id='bbb'></b></a></x>" 
      t.expand(out="", :outer=>TestPO.new) 
      assert_equal("<x><a><b>xxx</b><b>yyy</b><b>zzz</b></a></x>", out)

      t = TemplateText.new "<x id='outer'><a id='active'><b id='bbb'></b></a></x>" 
      t.expand(out="", :outer=>[TestPO.new, TestPO.new]) 
      assert_equal("<x><a><b>xxx</b><b>yyy</b><b>zzz</b></a></x><x><a><b>xxx</b><b>yyy</b><b>zzz</b></a></x>", out)

      t = TemplateText.new "<x id='outer'><a id='attr'><b id='bbb'></b></a></x>" 
      t.expand(out="", 'outer'=>TestPO.new) 
      assert_equal("<x><a aaa='bbb'><b yyy='zzz'>xxx</b></a></x>", out)
    end

    Lang = Struct.new(:name, :author, :website)
    class LangClass
      attr_reader :name, :author
      def initialize(n, a)
        @name, @author = n, a
      end
    end

    def test_template2
      tmpl = TemplateText.new <<-END
        <table>
          <tr><th>name</th><th>author</th></tr>
          <tr id='table1'>
            <td id='name'></td>
            <td id='author'></td>
          </tr>
       </table>
      END
      data = {
       :table1=>[ 
         Lang.new("Ruby", "matz"),
         LangClass.new("Perl", "Larry Wall"),
         Lang.new("Python","Guido van Rossum")
        ]
      }
      tmpl.expand(out="", data)
      formated_out = ""
      Open3.popen3("w3m -T text/html") do |i, o, e|
        i.puts out
        i.close
        formated_out = o.read
      end
      assert_equal(<<END, formated_out)
 name       author
Ruby   matz
Perl   Larry Wall
Python Guido van Rossum

END
    end

    def test_template_hash
      tmpl = TemplateText.new <<-END
        <table>
          <tr><th>name</th><th>author</th></tr>
          <tr id='table1'>
            <td id='name'></td>
            <td id='author'></td>
          </tr>
       </table>
      END
      data = {
       :table1=>[ 
         { :name=>"Ruby", :author=>'matz'},
         { :name=>"Perl", :author=>'Larry Wall'},
         { :name=>"Python", :author=>'Guido van Rossum'}
        ]
      }
      tmpl.expand(out="", data)
      formated_out = ""
      Open3.popen3("w3m -T text/html") do |i, o, e|
        i.puts out
        i.close
        formated_out = o.read
      end
      assert_equal(<<END, formated_out)
 name       author
Ruby   matz
Perl   Larry Wall
Python Guido van Rossum

END
    end

    class WebSite < Struct.new(:name, :url)
      def site(m)
        m.site(self.name, :href=>self.url)
      end
    end

    def test_template3
      tmpl = TemplateText.new <<-END
        <table>
          <tr><th>name</th><th>author</th><th>website</th></tr>
          <tr id='table1'>
            <td id='name'></td>
            <td id='author'></td>
            <td id='website'><a id='site'/></td>
          </tr>
       </table>
      END
      data = {
       "table1"=>[ 
         Lang.new("Ruby", "matz", WebSite.new('Ruby home page', 'http://www.ruby-lang.org')),
         Lang.new("Perl", "Larry Wall"),
         { 'name'=>"Python",'author'=>"Guido van Rossum" ,
           'website'=> WebSite.new('Python Programing Language', 'http://www.python.org/')},
        ]
      }
      tmpl.expand(out="", data)
      #print out
      formated_out = ""
      Open3.popen3("w3m -T text/html") do |i, o, e|
        i.puts out
        i.close
        formated_out = o.read
      end
      #print formated_out
      assert_equal(<<END, formated_out)
 name       author               website
Ruby   matz             Ruby home page
Perl   Larry Wall
Python Guido van Rossum Python Programing Language

END
    end

    def test_parts
      t = TemplateText.new '<x><span id="p1"><em id="aaa">xxx</em></span></x>' do 
        Amrita2::define_template_spec do 
          parts_element(:p1, :place_holder=>true) do 
            dynamic_element(:aaa)
          end
        end
      end
      t.setup      
      #assert_equal("<span id='p1'><em id='aaa'>123</em></span>", t.p1(:aaa=>'123'))
      assert_equal("<span id='p1'><em id='aaa'>123</em></span>", t.expand_parts(:p1,:aaa=>'123'))
      t.expand(out="") do
      end
      assert_equal("<x><span id='p1'></span></x>", out)
    end

    def test_parts2
      t = TemplateText.new '<x><span id="link"><a href="$1">$2</a></span></x>' do
        Amrita2::define_template_spec do 
          parts_element(:link, :use_args=>true)
        end
      end
      t.setup      
      assert_equal("<span><a href='http://amrita2.rubyforge.org/'>Amrita2</a></span>", t.link('http://amrita2.rubyforge.org/', 'Amrita2'))
    end

    def test_doctype
      t = TemplateText.new <<-END
<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" lang="en">
<head>
<title id="head_title" />
</head>
</html>
      END
      t.expand(out="", :head_title=>'test')
      assert_equal(<<-END, out)
<?xml version='1.0' encoding='UTF-8'?><!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html lang='en' xmlns='http://www.w3.org/1999/xhtml'>
<head>
<title>test</title>
</head>
</html>
END
     end
  end

  class TestOption < Test::Unit::TestCase
    include Amrita2

    def test_keep_id
      t = TemplateText.new "<a id='aaa'><b id='bbb'></b></a>" 
      t.keep_id = true
      t.expand(out="") do |m|
        m.aaa do |m2|
          m2.bbb('xxx')
        end
      end
      assert_equal("<a id='aaa'><b id='bbb'>xxx</b></a>", out)
    end

    def test_amrita_id
      t = TemplateText.new "<a aid='aaa'><b aid='bbb'></b></a>" 
      t.amrita_id = :aid
      t.expand(out="") do |m|
        m.aaa do |m2|
          m2.bbb('xxx')
        end
      end
      assert_equal("<a><b>xxx</b></a>", out)
    end

    def test_amrita_id2
      t = TemplateText.new "<a aid='aaa'><b aid='bbb'></b></a>" 
      t.amrita_id = :aid
      t.keep_id = true
      t.expand(out="") do |m|
        m.aaa do |m2|
          m2.bbb('xxx')
        end
      end
      assert_equal("<a aid='aaa'><b aid='bbb'>xxx</b></a>", out)
    end

    def test_ie_hack
      t = TemplateText.new  '<p><img src="foo.jpeg" alt="Foo" /><span id="aaa" /><img src="bar.jpeg" alt="Boo" /></p>'
      t.ie_hack = true
      t.expand(out="", :aaa=>'aaa')
      assert_equal("<p><img src='foo.jpeg' alt='Foo' /><span>aaa</span><img src='bar.jpeg' alt='Boo' /></p>", out)

      t = TemplateText.new  '<p><img src="foo.jpeg" alt="Foo" /><span id="aaa" /><img src="bar.jpeg" alt="Boo" /></p>'
      t.ie_hack = false
      t.expand(out="", :aaa=>'aaa')
      assert_equal("<p><img src='foo.jpeg' alt='Foo'/><span>aaa</span><img src='bar.jpeg' alt='Boo'/></p>", out)
    end
  end

  class TestInlineTDL < Test::Unit::TestCase
    include Amrita2
    include Amrita2::Runtime

    def assert_equal_xml(a, b)
      a = REXML::Document.new(a, :ignore_whitespace_nodes=>:all).to_s
      b = REXML::Document.new(b, :ignore_whitespace_nodes=>:all).to_s
      assert_equal(a, b)
    rescue 
      puts a
      puts 
      puts b
      raise
    end

    def test_use_args
      t = TemplateText.new "<span id='aaa' amrita:type='use_args'>$1/$2</span>" 
      t.expand(out="",:aaa=>['abc', 'efg'])
      assert_equal("<span>abc/efg</span>", out)
      t.expand(out="",:aaa=>[%w(aaa bbb), %w(xyz 789)])
      assert_equal("<span>aaa/bbb</span><span>xyz/789</span>", out)
    end

    def test_use_args_empty
      t = TemplateText.new "<span id='aaa' amrita:type='use_args'>$1/$2</span>" 
      t.expand(out="",:aaa=>[])
      assert_equal("", out)
      t.expand(out="",:aaa=>[%w(aaa bbb), %w(xyz 789), []])
      assert_equal("<span>aaa/bbb</span><span>xyz/789</span>", out)
    end

    def test_use_element
      t = TemplateText.new "<span id='aaa' amrita:type='use_original_element'>abc</span>" 
      p = proc do |e| 
        e.text += '789'; 
        e.attributes['class'] = 'xxx'
        e 
      end
      t.expand(out="",:aaa=> p)
      assert_equal("<span class='xxx'>abc789</span>", out)
    end

    def test_expand_by_member
      t = TemplateText.new "<span id='aaa' amrita:type='expand_by_member' amrita:option=':delete_tag=>true'><span id='year' /></span>" 
      t.expand(out="",:aaa=>Time.at(0))
      assert_equal("<span>1970</span>", out)
    end

    def test_value_attr
      t = TemplateText.new %[<span id='aaa' amrita:option=':value_attr=>"class"'>xxx</span>]
      t.expand(out="",:aaa=>'aaaa')
      assert_equal("<span class='aaaa'>xxx</span>", out)
    end

    def test_input
      t = TemplateText.new %[<input id='aaa' amrita:type='input' type='text' name='aaa[xxx]'/>]
      t.expand(out="",:aaa=>'aaaa')
      assert_equal("<input name='aaa[xxx]' type='text' value='aaaa'/>", out)
    end

    def test_select
      t = TemplateText.new %[<select id='sel1' amrita:type='select'><option value='a'>aaa</option><option value='b'>bbb</option><option value='c'>ccc</option> </select>]
      t.expand(out="",:sel1=>:a)
      assert_equal("<select><option selected='selected' value='a'>aaa</option><option value='b'>bbb</option><option value='c'>ccc</option></select>", out)
      t.expand(out="",:sel1=>:b)
      assert_equal("<select><option value='a'>aaa</option><option selected='selected' value='b'>bbb</option><option value='c'>ccc</option></select>", out)
      t.expand(out="",:sel1=>true)
      assert_equal("<select><option value='a'>aaa</option><option value='b'>bbb</option><option value='c'>ccc</option></select>", out)
    end

    def test_link
      t = TemplateText.new %[<a id='aaa' amrita:type='link' href='http://yyy/'>xxx</a>]
      t.expand(out="",:aaa=>%w(http://xxx/  wwww) )
      assert_equal("<a href='http://xxx/'>wwww</a>", out)
      t.expand(out="",:aaa=>a(:href=>'1111'))
      assert_equal("<a href='1111'>xxx</a>", out)
      t.expand(out="",:aaa=>a(:href=>'1111') { '333' })
      assert_equal("<a href='1111'>333</a>", out)
      t.expand(out="",:aaa=>[a(:href=>'1111') { '333' }, a(:ref=>'222'), %w(333, xxxx), '4444', a() { '555' }])
      assert_equal("<a href='1111'>333</a><a href='http://yyy/' ref='222'>xxx</a><a href='333,'>xxxx</a><a href='4444'>xxx</a><a href='http://yyy/'>555</a>", out)
    end

    def test_erb
      t = TemplateText.new %[<span id='aaa' amrita:type='erb'>aaa<![CDATA[<%= x * 2 %>]]>zzz</span>]
      x = 100
      t.expand(out="",:aaa=>binding)
      assert_equal("aaa200zzz", out)
    end

    def test_erb2
      t = TemplateText.new %[<span id='aaa' amrita:type='erb'><![CDATA[
        <ul>
        <% x.times do |i| %>
          <li><%= i * 2 %></li>
        <% end %>
        </ul>
      ]]></span>]
      x = 3
      t.expand(out="",:aaa=>binding)
      assert_equal_xml("<ul><li>0</li><li>2</li><li>4</li></ul>", out)
    end

    def test_erb3
      t = TemplateText.new %[<ul><span id='aaa' amrita:type='erb'><![CDATA[
          <li><%= x + $_ %></li>
      ]]></span></ul>]
      x = 'data:'
      b = binding
      data = %w(one two three).collect do |i|
        [b , i]
      end
      t.expand(out="",:aaa=>data)
      assert_equal_xml("<ul><li>data:one</li><li>data:two</li><li>data:three</li></ul>", out)
    end
  end
end

if __FILE__ == $0
  include Amrita2
  include Amrita2TemplateTest
  cname = ARGV.shift
  if cname
    method = ARGV.shift
    if method
      eval "Test::Unit::UI::Console::TestRunner.run(#{cname}.new(\"#{method}\"))"
    else
      cls = eval(cname)
      Test::Unit::UI::Console::TestRunner.run(cls)
    end
  else
    suite = Test::Unit::TestSuite.new
    suite << TestInlineTDL.suite
    suite << TestExpander.suite
    suite << TestOption.suite
    suite << TestTemplate.suite
    #suite << TestDefaultExpand.suite
    suite << TestDefaultSpec.suite
    Test::Unit::UI::Console::TestRunner.run(suite)
  end
end
